/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <omnetpp.h>

#include "api_m.h"
#include "slpv2_m.h"
#include "Service.h"
#include "util.h"
#include "debug.h"

// simple Client
//   parameters:
//     scopeList      : string,
//     serviceType    : string,
//     waitInterval   : numeric,
//     attrsRequired  : bool,
//     RFC3059Support : bool,
//     locale         : string;
// 
//   gates:
//      in: fromUA, fromNet;
//     out:   toUA,   toNet;
// endsimple

class Client : public cSimpleModule
{
 public:
  Module_Class_Members(Client, cSimpleModule, 0)

  const char* scopeList;
  const char* serviceType;
  bool attrsRequired;
  bool RFC3059Support;

  cArray serviceCache;
  SLPHandle slpHandle;
  SLPFindSrvs* slpFindSrvs;

  // Random selection policy
  bool serviceRequested;
  bool allServicesFound;
  int attributesFound;
  int servicesFound;
  char* serviceURLsFound[255];

  int id;

  cStdDev slpLatencyStatistics;
  cStdDev serviceLatencyStatistics;
  cStdDev errorStatistics;

  virtual void initialize();
  virtual void handleMessage(cMessage * msg);
  virtual void finish();

  virtual void handleFindSrvs
  (SLPFindSrvs * slp_find_srvs);

  virtual void handleSrvURLAttrsCallback
  (SLPSrvURLAttrsCallback* callback);

  virtual void handleSrvURLCallback
  (SLPSrvURLCallback * callback);
  
  virtual void handleAttrCallback
  (SLPAttrCallback* callback);

  virtual void handleSrvTypeCallback
  (SLPSrvTypeCallback * callback);

  virtual void requestService 
  ( const char* url, simtime_t tstamp );

  virtual void handleServerResponse
  ( ServerResponse* srv_rsp, Sockaddr_in* saddr );
};

Define_Module ( Client );

void Client::initialize()
{
#ifdef MOD_DEBUG
  debug(simTime(),parentModule()->fullName(), fullName(), 'D');
  ev.printf("initialize()\n");
#endif

  scopeList = par("scopeList").stringValue();
  if (strlen(scopeList) == 0) {
    scopeList = " ";
  }

  serviceType    = par("serviceType").stringValue();
  attrsRequired  = par("attrsRequired");
  RFC3059Support = attrsRequired && par("RFC3059Support");

  id = 0;

  serviceRequested = false;
  attributesFound = 0;
  allServicesFound = false;
  servicesFound = 0;
  for (int i=0; i<255; i++) {
    serviceURLsFound[i] = NULL;
  }

  serviceCache.setName("Service Cache");
  slpLatencyStatistics.setName("SLP Latency Statistics");
  errorStatistics.setName("Error Statistics");
  serviceLatencyStatistics.setName("Service Latency Statistics");

  slpHandle.pcLang = par("locale").stringValue();

  if (RFC3059Support) {
    slpFindSrvs = new SLPFindSrvsAttrs("SLPFindSrvsAttrs", SRV_RQST);
  } else {
    slpFindSrvs = new SLPFindSrvs("SLPFindSrvs", SRV_RQST);
  }
  slpFindSrvs->setHSLP(slpHandle);
  slpFindSrvs->setPcScopeList(scopeList);
  slpFindSrvs->setPcSearchFilter(" ");
  slpFindSrvs->setPcServiceType(serviceType);
    
  simtime_t wait_interval = par("waitInterval").doubleValue();
  scheduleAt(simTime() + wait_interval, slpFindSrvs);
}

void Client::handleMessage ( cMessage * msg )
{
#ifdef MOD_DEBUG
  debug(simTime(),parentModule()->fullName(), fullName(), 'D');
  ev.printf("handleMessage('%s')\n", msg->name());
#endif

  if (msg->isSelfMessage()) {

    handleFindSrvs(check_and_cast<SLPFindSrvs*>(msg));

  } else if (msg->arrivedOn("fromNet")) {

    Sockaddr_in* saddr = (Sockaddr_in*) msg->removeObject("recvFrom");
    handleServerResponse(check_and_cast<ServerResponse*>(msg), saddr);
    return;
  }

  SLPSrvURLAttrsCallback * srv_url_attrs_callback = 
    dynamic_cast<SLPSrvURLAttrsCallback*>(msg);
  if (srv_url_attrs_callback != NULL) {
    handleSrvURLAttrsCallback(srv_url_attrs_callback);
    return;
  }

  SLPSrvURLCallback * srv_url_callback = 
    dynamic_cast<SLPSrvURLCallback*>(msg);
  if (srv_url_callback != NULL) {
    handleSrvURLCallback(srv_url_callback);
    return;
  }

  SLPSrvTypeCallback * srv_type_callback = 
    dynamic_cast<SLPSrvTypeCallback*>(msg);
  if (srv_type_callback != NULL) {
    handleSrvTypeCallback(srv_type_callback);
    return;
  }

  SLPAttrCallback * attr_callback = 
    dynamic_cast<SLPAttrCallback*>(msg);
  if (attr_callback != NULL) {
    handleAttrCallback(attr_callback);
    return;
  }
}

void Client::handleFindSrvs ( SLPFindSrvs * slp_find_srvs )
{
  // Handle periodic SLPFindSrvs
  //

  if (strlen(serviceType) == 0) {

    // No service type has been configured, send a SLPFindSrvsTypes
    //
    SLPFindSrvsTypes* slp_find_srvs_types =
      new SLPFindSrvsTypes("SLPFindSrvsTypes", SRV_TYPE_RQST);
    slp_find_srvs_types->setHSLP(slpHandle);
    slp_find_srvs_types->setPcScopeList(scopeList);

#ifdef API_DEBUG
    debug(simTime(), parentModule()->fullName(), fullName(), '>');
    debug(slp_find_srvs_types);
#endif

    send(slp_find_srvs_types, "toUA");
    return;
  }

  // Look for active services in the cache
  for (int i=0; i<serviceCache.items(); i++) {
    Service * service = (Service*) serviceCache[i];
    if (service == NULL) {
      continue;
    }
    if (service->hasExpired(simTime())) {
      delete serviceCache.remove(i);
      continue;
    }

    if (service->hasType(serviceType)) {

#ifdef API_DEBUG
      if (servicesFound == 0) {
	debug(simTime(), parentModule()->fullName(), fullName(), ':');
	ev << "'" << serviceType << "' service(s) found in cache:" 
	   << endl;
      }
      debug('|');
      ev.printf("%.25s (", service->getURL());
      ev << (int) service->getTTL(simTime()) << "/" 
	 << service->getLifetime() << ")  ";
      ev.printf("%.30s\n", service->getAttributes());
#endif

      serviceURLsFound[servicesFound] = strdup(service->getURL());
      servicesFound++;
    }
  }

  if (servicesFound > 0) {

    allServicesFound = true;
    int index = (int) uniform(0.0, (double) servicesFound);
    requestService(serviceURLsFound[index], simTime());

  } else {

    // No services have been found in cache: Call to SLPFindSrvs()
    //
    SLPFindSrvs * slp_find_srvs = (SLPFindSrvs*) slpFindSrvs->dup();
    slp_find_srvs->setTimestamp(simTime());
    
#ifdef API_DEBUG
    debug(simTime(), parentModule()->fullName(), fullName(), '>');
    debug(slp_find_srvs);
#endif
    
    send(slp_find_srvs, "toUA");
  }
}


void Client::handleSrvURLAttrsCallback ( SLPSrvURLAttrsCallback * callback )
{

#ifdef API_DEBUG
  debug(simTime(),parentModule()->fullName(), fullName(), '<');
  debug(callback);
#endif

  if (callback->getErrCode() == SLP_NETWORK_TIMED_OUT) {
    // Update error statistics
    simtime_t latency = simTime() - callback->timestamp();
    errorStatistics.collect(latency);
    
    // Re-schedule periodic slpFindSrvs
    if (! slpFindSrvs->isScheduled()) {
      simtime_t wait_interval = par("waitInterval").doubleValue();
      scheduleAt(simTime() + wait_interval, slpFindSrvs);
    }

    delete callback;
    return;
  }

  const char * url = callback->getPcSrvURL();
  int lifetime = callback->getSLifetime();
  const char * attr_list = callback->getPcAttrList();

  if (strcmp(url, " ") == 0) {

    if (servicesFound > 0) {

      // Multicast Service Discovery has ended
      int index = (int) uniform(0.0, (double) servicesFound);
      requestService(serviceURLsFound[index], callback->timestamp());

    } else if (! slpFindSrvs->isScheduled()) {

      // Empty DA response: Re-schedule slpFindSrvs
      simtime_t wait_interval = par("waitInterval").doubleValue();
      scheduleAt(simTime() + wait_interval, slpFindSrvs);
    }

    delete callback;
    return;
  }

  // Update cache
  Service * srv = (Service *) serviceCache[url];
  if (srv == NULL) {
    // New service: add it to the cache
    srv = new Service(serviceType, url, attr_list, lifetime, 
		      simTime());
    serviceCache.add(srv);
  } else {
    // Service already in cache: refresh it
    srv->refresh(simTime());
  }

  serviceURLsFound[servicesFound] = strdup(url);
  servicesFound++;

  if (callback->getErrCode() == SLP_LAST_CALL) {
    allServicesFound = true;
    int index = (int) uniform(0.0, (double) servicesFound);
    requestService(serviceURLsFound[index], callback->timestamp());
  }

  delete callback;
}

void Client::handleSrvURLCallback ( SLPSrvURLCallback * callback )
{
#ifdef API_DEBUG
    debug(simTime(), parentModule()->fullName(), fullName(), '<');
    debug(callback);
#endif
    
  if (callback->getErrCode() == SLP_NETWORK_TIMED_OUT) {
    // Update error statistics
    simtime_t latency = simTime() - callback->timestamp();
    errorStatistics.collect(latency);
    
    // Re-schedule periodic slpFindSrvs
    if (! slpFindSrvs->isScheduled()) {
      simtime_t wait_interval = par("waitInterval").doubleValue();
      scheduleAt(simTime() + wait_interval, slpFindSrvs);
    }

    delete callback;
    return;
  }

  const char* url = callback->getPcSrvURL();
  int lifetime = callback->getSLifetime();

  if (strcmp(url, " ") == 0) {

    if (servicesFound > 0) {

      // Multicast Service Discovery has ended
      allServicesFound = true;
      if ((! attrsRequired) || (servicesFound == attributesFound)) {
	int index = (int) uniform(0.0, (double) servicesFound);
	requestService(serviceURLsFound[index], callback->timestamp());
      }

    } else if (! slpFindSrvs->isScheduled()) {

      // Empty DA response: Re-schedule slpFindSrvs
      simtime_t wait_interval = par("waitInterval").doubleValue();
      scheduleAt(simTime() + wait_interval, slpFindSrvs);
    }

    delete callback;
    return;
  }

  // Update cache
  Service * srv = (Service *) serviceCache[url];
  if (srv == NULL) {
    // New service: add it to the cache
    srv = new Service(serviceType, url, lifetime, simTime());
    serviceCache.add(srv);
  } else {
    // Service already in cache: refresh it
    srv->refresh(simTime());
  }

  serviceURLsFound[servicesFound] = strdup(url);
  servicesFound++;

  if (attrsRequired) {

    if (callback->getErrCode() == SLP_LAST_CALL) {
      allServicesFound = true;
    }

    // Service discovery requires also the attributes also: Call
    // SLPFindAttrs()

    SLPFindAttrs * slp_find_attrs = new SLPFindAttrs
      ("SLPFindAttrs", ATTR_RQST);
    slp_find_attrs->setHSLP(slpHandle);
    slp_find_attrs->setPcURL(url);
    slp_find_attrs->setPcScopeList(scopeList);
    slp_find_attrs->setPcAttrIds(" ");
    slp_find_attrs->setContextPointer(callback);
    slp_find_attrs->setTimestamp(callback->timestamp());

#ifdef API_DEBUG
    debug('>');
    debug(slp_find_attrs);
#endif

    send(slp_find_attrs, "toUA");

  } else {

    // Attributes are not needed, service discovery is over
    if (callback->getErrCode() == SLP_LAST_CALL) {
      allServicesFound = true;
      int index = (int) uniform(0.0, (double) servicesFound);
      requestService(serviceURLsFound[index], callback->timestamp());
    }

    delete callback;
  }
}

void Client::handleAttrCallback ( SLPAttrCallback * callback )
{
#ifdef API_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(callback);
#endif

  SLPSrvURLCallback * srv_url_callback = 
    (SLPSrvURLCallback *) callback->contextPointer();
  const char* url = srv_url_callback->getPcSrvURL();

  if (callback->getErrCode() == SLP_NETWORK_TIMED_OUT) {

    // Update Error Statistics
    simtime_t latency = simTime() - callback->timestamp();
    errorStatistics.collect(latency);
    
    delete serviceCache.remove(url);
    servicesFound--;
    free(serviceURLsFound[servicesFound]);

    // Re-schedule periodic slpFindSrvs
    if (! slpFindSrvs->isScheduled()) {
      simtime_t wait_interval = par("waitInterval").doubleValue();
      scheduleAt(simTime() + wait_interval, slpFindSrvs);
    }

    delete callback;
    return;
  }

  // Update cache
  Service * srv = (Service *) serviceCache[url];
  if (srv != NULL) {
    const char* attr_list = callback->getPcAttrList();
    srv->setAttributes(attr_list);
  }

  attributesFound++;
  if ((allServicesFound) && (attributesFound == servicesFound)) {
    int index = (int) uniform(0.0, (double) servicesFound);
    requestService(serviceURLsFound[index], callback->timestamp());
  }

  delete srv_url_callback;
  delete callback;
}

void Client::handleSrvTypeCallback ( SLPSrvTypeCallback * callback )
{
#ifdef API_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(callback);
#endif
    
  if ((callback->getErrCode() != SLP_NETWORK_TIMED_OUT) && 
      (strlen(serviceType) == 0)) {
      char* types = (char*) callback->getPcSrvTypes();
      serviceType = strdup(strsep(&types, ","));
      par("serviceType").setStringValue(serviceType);
    }

  // Schedule periodic slpFindSrvs
  if (! slpFindSrvs->isScheduled()) {
    slpFindSrvs->setPcServiceType(serviceType);
    simtime_t wait_interval = par("waitInterval").doubleValue();
    scheduleAt(simTime() + wait_interval, slpFindSrvs);
  }  
  
  delete callback;
}

void Client::requestService ( const char* url, simtime_t tstamp )
{
  if (serviceRequested) {
    return;
  } else {
    serviceRequested = true;
  }

  // Update SLP Statistics
  simtime_t latency = simTime() - tstamp;
  slpLatencyStatistics.collect(latency);

  char* srv_addr = url2addr(url);

  ClientRequest * clnt_rqst = new ClientRequest("ClntRqst", 12);
  clnt_rqst->setTimestamp(tstamp);
  clnt_rqst->setId(id++);

  int bytes = 4; // Id
  clnt_rqst->setLength(bytes);

  Sockaddr_in* saddr = new Sockaddr_in("sendTo", "server", srv_addr);
  clnt_rqst->addObject(saddr);

#ifdef CLIENT_SERVER_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>'); 
  debug(clnt_rqst, saddr);
#endif
    
  send(clnt_rqst, "toNet");
}

void Client::handleServerResponse
( ServerResponse * srv_rsp, Sockaddr_in* saddr )
{
#ifdef CLIENT_SERVER_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(srv_rsp, saddr);
#endif  

  // Update Service Statistics
  simtime_t latency = simTime() - srv_rsp->timestamp();
  serviceLatencyStatistics.collect(latency);

  // Re-schedule periodic slpFindSrvs
  if (! slpFindSrvs->isScheduled()) {
    simtime_t wait_interval = par("waitInterval").doubleValue();
    scheduleAt(simTime() + wait_interval, slpFindSrvs);
  }

  serviceRequested = false;
  attributesFound = 0;
  allServicesFound = false;
  for (int i=0; i<servicesFound; i++) {
    free(serviceURLsFound[i]);
  }
  servicesFound = 0;

  delete srv_rsp;
  delete saddr;
}


void Client::finish()
{
  ev << endl << "------- Client " << parentModule()->fullName()
     << " statistics -------" << endl;

  ev << "Options:";
  if (attrsRequired)  ev << "  (Attrs Required)";
  if (RFC3059Support) ev << "  (RFC3059 Support)";
  ev << endl << endl;

  ev << "Services discovered : " 
     << slpLatencyStatistics.samples() << endl;
  ev << "   Latency : "
     << slpLatencyStatistics.min() * 1000.0 << "/"
     << slpLatencyStatistics.mean() * 1000.0 << "/"
     << slpLatencyStatistics.max() * 1000.0 << "/"
     << slpLatencyStatistics.stddev() * 1000.0 << " ms"
     << endl;

  if (errorStatistics.samples() > 0) {
    ev << endl;
    ev << "Service discovery errors : " 
       << errorStatistics.samples() << " !!!"<< endl;
    ev << "   Latency : "
       << errorStatistics.min() * 1000.0 << "/"
       << errorStatistics.mean() * 1000.0 << "/"
       << errorStatistics.max() * 1000.0 << "/"
       << errorStatistics.stddev() * 1000.0 << " ms"
       << endl;
  }

  ev << "Services requested : " 
     << serviceLatencyStatistics.samples() << endl;
  ev << "   Latency : "
     << serviceLatencyStatistics.min() * 1000.0 << "/"
     << serviceLatencyStatistics.mean() * 1000.0 << "/"
     << serviceLatencyStatistics.max() * 1000.0 << "/"
     << serviceLatencyStatistics.stddev() * 1000.0 << " ms"
     << endl;
}
