/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <omnetpp.h>
#include <string.h>

#include "ServiceAgent.h"
#include "slpv2_m.h"
#include "Service.h"
#include "Sockaddr_in.h"
#include "util.h"
#include "debug.h"

// simple DirectoryAgent
//   parameters:
//     DAHeartBeat      : numeric,
//     DAAttributes     : string,
//     useScopes        : string,
//     MTU              : numeric;
// 
//   gates:
//      in: fromNet;
//     out:   toNet;
// endsimple

class DirectoryAgent : public ServiceAgent
{
  Module_Class_Members ( DirectoryAgent, ServiceAgent, 0 )

  const char* useScopes;

  DAAdvertisement* daAdvertisement;

  virtual void initialize();
  virtual void handleMessage(cMessage *msg);

  virtual void handleDAAdvertisement
  (DAAdvertisement* da_advert);

  virtual void handleServiceRegistration
  (ServiceRegistration* srv_reg, Sockaddr_in* saddr);

  virtual SLPv2Message* createAdvertisementReply
  ( ServiceRequest* srv_rqst );
};

Define_Module ( DirectoryAgent )

void DirectoryAgent::initialize()
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("initialize()\n");
#endif

  useScopes = par("useScopes").stringValue();
  MTU = par("MTU");

  const char* hostname = parentModule()->fullName();
  int url_len = 27 + strlen(hostname);
  char* da_url = (char*) malloc(url_len);
  sprintf(da_url, "service:directory-agent://%s", hostname);

  char* attrs = (char*) par("DAAttributes").stringValue();
  char* da_attrs = expandVar(attrs, "$HOST", hostname);
  par("DAAttributes").setStringValue(attrs);
  //free(attrs);

  agentService = new Service("service:directory-agent", da_url, 
			     da_attrs, -1, simTime());
  serviceCache.add(agentService);
  serviceCache.setName("Service Cache");

  timers.setName("Timers");
  slpQueueLength.setName("QueueLength");

  daAdvertisement = new DAAdvertisement("DAAdvert", DA_ADVERT);
  daAdvertisement->setVersion(2);
  daAdvertisement->setFunctionId(DA_ADVERT);
  daAdvertisement->setOverflow(false);
  daAdvertisement->setFresh(false);
  daAdvertisement->setRequestMulticast(false);
  daAdvertisement->setXid(0);
  daAdvertisement->setLanguageTag("en");
  daAdvertisement->setErrorCode(OK);
  daAdvertisement->setBootTimestamp((long) simTime()*1000);
  daAdvertisement->setURL(da_url);
  daAdvertisement->setScopeList(useScopes);
  daAdvertisement->setAttrList(da_attrs);

  int bytes = 
    16 +                      // SLPv2 header: "en" language tag
     2 +                      // Error code
     4 +                      // DA Boot Timestamp
     2 + strlen(da_url) +     // URL
     2 + strlen(useScopes) +  // Scope list
     2 + strlen(da_attrs)+    // Attribute list
     3;                       // SPI Length + #Auth Blocks=0
 
  daAdvertisement->setLength(bytes*8);

  simtime_t da_heart_beat = par("DAHeartBeat").doubleValue();
  if (da_heart_beat > 0.0) {
    scheduleAt(0, daAdvertisement);
  }
}


void DirectoryAgent::handleMessage( cMessage *msg )
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("handleMessage('%s()')\n", msg->name());
#endif

  if (msg->isSelfMessage()) {

    handleDAAdvertisement 
      (check_and_cast<DAAdvertisement*>(msg));

  } else {

    Sockaddr_in* saddr = (Sockaddr_in*) msg->removeObject("recvFrom");

    SLPv2Message * slp = check_and_cast<SLPv2Message*>(msg);
    switch(slp->getFunctionId()) {
    case SRV_REG:
      handleServiceRegistration
	(check_and_cast<ServiceRegistration*>(slp), saddr);
      break;
    case SRV_RQST:
      handleServiceRequest
	(check_and_cast<ServiceRequest*>(slp), saddr);
      break;
    case SRV_TYPE_RQST:
      handleServiceTypeRequest
	(check_and_cast<ServiceTypeRequest *>(slp), saddr);
      break;
    case ATTR_RQST:
      handleAttributeRequest
	(check_and_cast<AttributeRequest *>(slp), saddr);
      break;
    }
  }
}


void DirectoryAgent::handleDAAdvertisement
( DAAdvertisement* da_advert ) 
{
  Sockaddr_in* saddr = new Sockaddr_in("sendTo", "slp", "*");
   
  DAAdvertisement * copy = (DAAdvertisement*) da_advert->dup();
  copy->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>');
  debug(da_advert, saddr);
#endif

  send(copy, "toNet");
  
  simtime_t da_heart_beat = par("DAHeartBeat").doubleValue();
  scheduleAt(simTime() + da_heart_beat, da_advert);
}


void DirectoryAgent::handleServiceRegistration 
( ServiceRegistration* srv_reg, Sockaddr_in* saddr )
{
  URLEntry url_entry = srv_reg->getEntry();
  const char* url = url_entry.URL;
  int lifetime = url_entry.lifetime;
  const char* service_type = srv_reg->getServiceType();
  const char* scopes = srv_reg->getScopeList();
  const char* attrs = srv_reg->getAttrList();

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(srv_reg, saddr);
#endif

  // Update cache
  Service * srv = (Service *) serviceCache[url];
  if (srv == NULL) {
    // New service: add it to the cache
    srv = new Service(service_type, url, attrs, lifetime, simTime());
    serviceCache.add(srv);
  } else {
    // Service already in cache: refresh it
    if (srv_reg->getFresh()) {
      srv->setAttributes(attrs);
    }
    srv->refresh(simTime());
  }

  ServiceAcknowledgement * srv_ack 
    = new ServiceAcknowledgement("SrvAck", SRV_ACK);
  srv_ack->setTimestamp(srv_reg->timestamp());
  srv_ack->setContextPointer(srv_reg->contextPointer());
  srv_ack->setVersion(2);
  srv_ack->setFunctionId(SRV_ACK);
  srv_ack->setOverflow(false);
  srv_ack->setFresh(false);
  srv_ack->setRequestMulticast(false);
  srv_ack->setXid(srv_reg->getXid());
  srv_ack->setLanguageTag("en");
  srv_ack->setErrorCode(OK);

  int bytes =
    16 +       // SLPv2 header: 2 letter lang tag
     2;        // Error code

  srv_ack->setLength(bytes*8);

  saddr->setName("sendTo");
  srv_ack->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug('>');
  debug(srv_ack, saddr);
#endif

  sendReply(srv_ack);
  delete srv_reg;
}


SLPv2Message* DirectoryAgent::createAdvertisementReply
( ServiceRequest* srv_rqst )
{
  // Active DA Discovery
  //
  DAAdvertisement* da_advert = 
    (DAAdvertisement*) daAdvertisement->dup();
  da_advert->setTimestamp(srv_rqst->timestamp());
  da_advert->setContextPointer(srv_rqst->contextPointer());
  da_advert->setXid(srv_rqst->getXid());
  
  return da_advert;
}
