/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <omnetpp.h>
#include <string.h>

#include "network_m.h"
#include "debug.h"

// simple Network
//   gates:
//     in: portIn[]
//     out: portOut[]
// endsimple

class Network : public cSimpleModule 
{
 public:
  Module_Class_Members(Network, cSimpleModule, 0)

  int numOfPorts;

  cStdDev errorStats;
  cStdDev frameStats;
  cStdDev unicastStats;
  cStdDev broadcastStats;

  virtual void initialize();
  virtual void handleMessage( cMessage *msg );
  virtual void finish();
};

Define_Module ( Network );

void Network::initialize()
{
#ifdef MOD_DEBUG
  debug(simTime(), fullName(), NULL, 'D');
  ev.printf("initialize()\n");
#endif

  numOfPorts = gate("portOut")->size();

  errorStats.setName("errorStats");
  frameStats.setName("frameStats");
  unicastStats.setName("unicastStats");
  broadcastStats.setName("broadcastStats");
}

void Network::handleMessage( cMessage *msg )
{
#ifdef MOD_DEBUG
  debug(simTime(), fullName(), NULL, 'D');
  ev.printf("handleMessage('%s()')\n", msg->name());
#endif

  IPDatagram * dgram = check_and_cast<IPDatagram*>(msg);

  if (dgram->hasBitError()) {

#ifdef NET_DEBUG
    debug(simTime(), fullName(), dgram->getSourceAddress(), 'X');
    ev << dgram->name() << "("<< dgram->length()/8 << " bytes) -> " 
       << dgram->getDestinationAddress() << "  Dropped !!" << endl;
#endif  

    errorStats.collect(dgram->length()/8);
    delete dgram;
    return;
  }

  const char * dest_addr = dgram->getDestinationAddress();

#ifdef NET_DEBUG
  debug(simTime(), fullName(), dgram->getSourceAddress(), '<');
  ev << dgram->name() << "(" << dgram->length()/8 << " bytes) -> "
     << dgram->getDestinationAddress() << endl;
#endif  
  
  if (strcmp(dest_addr, "*") == 0) {
    // Broadcast forwarding
    //
    broadcastStats.collect(dgram->length()/8);

    for (int i=0; i<numOfPorts; i++) {
      if (dgram->arrivedOn("portIn", i)) {
	continue;
      }

#ifdef NET_DEBUG
      cGate * out_gate = gate("portOut", i)->destinationGate();
      cModule * out_netif = out_gate->ownerModule();
      cModule * out_host = out_netif->parentModule();
      const char* out_addr = out_host->fullName();
      
      debug('>');
      ev << dgram->name() << "() -> " << out_addr << endl;
#endif  
	
      IPDatagram * copy = (IPDatagram *) dgram->dup();
      send(copy, "portOut", i);
      frameStats.collect(copy->length()/8);
    }
    delete dgram;

  } else {
    // Unicast forwarding
    //
    unicastStats.collect(dgram->length()/8);

    for (int i=0; i<numOfPorts; i++) {
      cGate * out_gate = gate("portOut", i)->destinationGate();
      cModule * out_netif = out_gate->ownerModule();
      cModule * out_host = out_netif->parentModule();
      const char* out_addr = out_host->fullName();

      if (strcmp(dest_addr, out_addr) == 0) {
  	send(dgram, "portOut", i);
	frameStats.collect(dgram->length()/8);
	break;
      }
    }
  }
}

void Network::finish()
{
  ev.printf("\n------- Network statistics -------\n");
  ev << "  Unicast packets : " << unicastStats.samples()
     << " (" << unicastStats.sum() << " bytes) " 
     << unicastStats.min() << "/" 
     << unicastStats.mean() << "/"
     << unicastStats.max() << "/"  
     << unicastStats.stddev() << " bytes" << endl;
  ev << "Broadcast packets : " << broadcastStats.samples()
     << " (" << broadcastStats.sum() << " bytes) " 
     << broadcastStats.min() << "/" 
     << broadcastStats.mean() << "/"
     << broadcastStats.max() << "/" 
     << broadcastStats.stddev() << " bytes" << endl;
  ev << "     Total frames : " << frameStats.samples()
     << " (" << frameStats.sum() << " bytes) " 
     << frameStats.min() << "/" 
     << frameStats.mean() << "/"
     << frameStats.max() << "/" 
     << frameStats.stddev() << " bytes" << endl;

  ev << endl << "Total simulation time: " << simtimeToStr(simTime()) 
     << endl;
}
