/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <stdio.h>
#include <time.h>
#include <string.h>

#include "network_m.h"
#include "Sockaddr_in.h"
#include "debug.h"

// simple NetworkInterface
// 
//   gates:
//      in: fromApps[], fromPort;
//     out:   toApps[], toPort;
// endsimple

class NetworkInterface : public cSimpleModule 
{
 public:
  Module_Class_Members(NetworkInterface, cSimpleModule, 0)

  int numOfApps;

  cStdDev input_error_stats;
  cStdDev input_unknown_stats;
  cStdDev input_packet_stats;
  cStdDev output_packet_stats;

  virtual void initialize();
  virtual void handleMessage( cMessage *msg );
};

Define_Module ( NetworkInterface );

void NetworkInterface::initialize()
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("initialize()\n");
#endif

  numOfApps = gate("toApps")->size();

  input_error_stats.setName("input_error_stats");
  input_unknown_stats.setName("input_unknown_stats");
  input_packet_stats.setName("input_packet_stats");
  output_packet_stats.setName("output_packet_stats");
}

void NetworkInterface::handleMessage( cMessage *msg )
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("handleMessage('%s()')\n", msg->name());
#endif
  
  if (! msg->arrivedOn("fromPort")) {

    UDPDatagram * dgram = new UDPDatagram(msg->name(), msg->kind());
    dgram->setLength(28*8);
    dgram->encapsulate(msg);

    cGate* in_gate = msg->arrivalGate()->sourceGate();
    cModule* in_app = in_gate->ownerModule();
    dgram->setSourcePort(in_app->fullName());
    dgram->setSourceAddress(parentModule()->fullName());

    Sockaddr_in* saddr = (Sockaddr_in*) msg->removeObject("sendTo");
    if (saddr == NULL) {
      error("Message from application '%s' has not 'sendTo' struct",
	    in_app->fullName());
    }
    dgram->setDestinationAddress(saddr->getAddress());
    dgram->setDestinationPort(saddr->getPort());

#ifdef NET_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>');
  ev << dgram->name() << "(" << dgram->length()/8 << " bytes) -> "
     << dgram->getDestinationAddress() << endl;
#endif

    send(dgram, "toPort");

    output_packet_stats.collect(dgram->length()/8);

    delete saddr;

  } else {
    UDPDatagram *dgram = check_and_cast<UDPDatagram*>(msg);

    if (dgram->hasBitError()) {
    
#ifdef NET_DEBUG
      debug(simTime(), parentModule()->fullName(), fullName(), 'X');
      ev << dgram->name() << "("<< dgram->length()/8 << " bytes) <- "
	 << dgram->getSourceAddress() << "  Dropped !!" << endl;
#endif
      
      input_error_stats.collect(dgram->length()/8);
      delete dgram;
      return;
    }
      
#ifdef NET_DEBUG
    debug(simTime(), parentModule()->fullName(), fullName(), '<');
    ev << dgram->name() << "(" << dgram->length()/8 << " bytes) <- "
       << dgram->getSourceAddress() << endl;
#endif
    
    const char * dest_port = dgram->getDestinationPort();
    
    for (int i=0; i<numOfApps; i++) {
      cGate   * out_gate = gate("toApps", i)->destinationGate();
      cModule * out_app = out_gate->ownerModule();
      const char* app_name = out_app->fullName();
      
      if (strcmp(dest_port, app_name) == 0) {
	
	cMessage * msg = dgram->decapsulate();
	
	Sockaddr_in* saddr = new Sockaddr_in
	  ("recvFrom", dgram->getSourcePort(), 
	   dgram->getSourceAddress());
	msg->addObject(saddr);
	
	send(msg, "toApps", i);
	input_packet_stats.collect(dgram->length()/8);
	
	delete dgram;
	return;
      }
    }
    
    input_unknown_stats.collect(dgram->length()/8);
    delete dgram;
  }
}
