/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <omnetpp.h>

#include "api_m.h"
#include "slpv2_m.h"
#include "client_server_m.h"
#include "util.h"
#include "debug.h"

// simple Client
//   parameters:
//     serviceType    : string,
//     waitInterval   : numeric,
//     attrsRequired  : bool,
//     RFC3059Support : bool,
//     locale         : string;
// 
//   gates:
//      in: fromUA, fromNet;
//     out:   toUA,   toNet;
// endsimple

class Server : public cSimpleModule
{
  const char* serviceType;
  const char* serviceURL;
  const char* serviceAttrs;
  int serviceLifetime;

  simtime_t occupiedUntil;
  cOutVector queueLength;

  cStdDev serviceStatistics;

  SLPHandle slpHandle;
  SLPReg* slpReg;
  
  Module_Class_Members ( Server, cSimpleModule, 0 )

  virtual void initialize();
  virtual void handleMessage(cMessage* msg);
  virtual void finish();

  virtual void handleRegReport(SLPRegReport* slp_reg_report);

  virtual void handleClientRequest
  ( ClientRequest* clnt_rqst, Sockaddr_in* saddr );
};

Define_Module ( Server )

void Server::initialize()
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("initialize()\n");
#endif

  const char* var = "$HOST";
  const char* val = parentModule()->fullName();

  char* url = (char*) par("serviceURL").stringValue();
  url = expandVar(url, var, val);
  serviceURL = url;
  par("serviceURL").setStringValue(serviceURL);

  char* attrs = (char*) par("serviceAttrs").stringValue();
  attrs = expandVar(attrs, var, val);
  serviceAttrs = attrs;
  par("serviceAttrs").setStringValue(serviceAttrs);

  serviceType = par("serviceType").stringValue();
  serviceLifetime = par("serviceLifetime");

  slpHandle.pcLang = par("locale").stringValue();

  queueLength.setName("QueueLength");
  serviceStatistics.setName("Service Statistics");
  
  slpReg = new SLPReg("SLPReg", SRV_REG);
  slpReg->setHSLP(slpHandle);
  slpReg->setPcSrvURL(serviceURL);
  slpReg->setUsLifetime(serviceLifetime);
  slpReg->setPcSrvType(serviceType);
  slpReg->setPcAttrs(serviceAttrs);
  slpReg->setFresh(true);

  simtime_t init_delay = uniform(1.5, 3.5);
  scheduleAt(init_delay, slpReg);
}


void Server::handleMessage ( cMessage *msg )
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("handleMessage('%s')\n", msg->name());
#endif

  if (msg->isSelfMessage()) {
    SLPReg * slp_reg = check_and_cast<SLPReg*>(msg);
    
    if (slpReg->getFresh() == false) {
      int workload = 0;
      if (occupiedUntil > simTime()) {
	workload = (int) (occupiedUntil - simTime());
      }

      char attrs[64];
      sprintf(attrs, "(service-workload=%d)", workload);
      slpReg->setPcAttrs(attrs);
    }

#ifdef API_DEBUG
    debug(simTime(), parentModule()->fullName(), fullName(), '>');
    ev.printf("%s('%.25s' (%d), '%s', '%.30s'", slp_reg->name(), 
	      slp_reg->getPcSrvURL(), slp_reg->getUsLifetime(), 
	      slp_reg->getPcSrvType(), slp_reg->getPcAttrs());
    if (slp_reg->getFresh()) {
      ev << ", F";
    }
    ev << ")" << endl;
#endif

    SLPReg * copy = (SLPReg*) slp_reg->dup();
    send(copy, "toSA");

    slpReg->setFresh(false);
    return;

  } else if (msg->arrivedOn("fromNet")) {

    Sockaddr_in* saddr = (Sockaddr_in*) msg->removeObject("recvFrom");
    handleClientRequest(check_and_cast<ClientRequest*>(msg), saddr);
    return;
  }

  SLPRegReport * slp_reg_report = dynamic_cast<SLPRegReport *>(msg);
  if (slp_reg_report != NULL) {
    handleRegReport(slp_reg_report);
    return;
  }

  error("Unknown message received: '%s'", msg->name());
}

void Server::handleRegReport ( SLPRegReport* slp_reg_report )
{

#ifdef API_DEBUG
    debug(simTime(), parentModule()->fullName(), fullName(), '<');
    ev << slp_reg_report->name() << "(";
    if (slp_reg_report->getErrCode() == SLP_OK) {
      ev << "SLP_OK";
    } else {
      ev << slp_reg_report->getErrCode();
    }
    ev << ")" <<  endl;
#endif  

    delete slp_reg_report;

    simtime_t wait_interval = par("waitInterval").doubleValue();
    scheduleAt(simTime() + wait_interval, slpReg);
}

void Server::handleClientRequest
( ClientRequest * clnt_rqst, Sockaddr_in* saddr )
{

#ifdef CLIENT_SERVER_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(clnt_rqst, saddr);
#endif

  ServerResponse * srv_rsp = new ServerResponse("SrvRsp", 8);
  srv_rsp->setTimestamp(clnt_rqst->timestamp());
  srv_rsp->setContextPointer(clnt_rqst->contextPointer());
  srv_rsp->setId(clnt_rqst->getId());

  int bytes = 4; // Id
  srv_rsp->setLength(bytes);

  saddr->setName("sendTo");
  srv_rsp->addObject(saddr);

#ifdef CLIENT_SERVER_DEBUG
  debug('>');
  debug(srv_rsp, saddr);
#endif  

  simtime_t processing_time = par("processingTime").doubleValue();
  if (occupiedUntil > simTime()) {
    simtime_t queue_time = occupiedUntil - simTime();
    queueLength.record(queue_time);
    occupiedUntil += processing_time;
    processing_time += queue_time;
  } else {
    queueLength.record(0.0);
    occupiedUntil = simTime() + processing_time;
  }

  serviceStatistics.collect(processing_time);

  sendDelayed(srv_rsp, processing_time, "toNet");

  delete clnt_rqst;
}

void Server::finish()
{
  ev << endl << "------- Server " << parentModule()->fullName()
     << " statistics -------" << endl;

  ev << "Services requested : " 
     << serviceStatistics.samples() << endl;
  ev << "   Response Time : "
     << serviceStatistics.min() * 1000.0 << "/"
     << serviceStatistics.mean() * 1000.0 << "/"
     << serviceStatistics.max() * 1000.0 << "/"
     << serviceStatistics.stddev() * 1000.0 << " ms"
     << endl;
}
