/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <omnetpp.h>

#include "ServiceAgent.h"
#include "Service.h"
#include "Sockaddr_in.h"
#include "util.h"
#include "debug.h"

// simple ServiceAgent
//   parameters:
//     useScopes          : string,
//     DAAddresses        : string,
//     datagramTimeout    : numeric,
//     ramdomWaitBound    : numeric,
//     MTU                : numeric,
//     SAAttributes       : string;
// 
//   gates:
//      in: fromApp, fromNet;
//     out:   toApp,   toNet;
// endsimple


Define_Module ( ServiceAgent )

void ServiceAgent::initialize()
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("initialize()\n");
#endif

  UserAgent::initialize();

  const char* hostname = parentModule()->fullName();
  int url_len = 27 + strlen(hostname);
  char* sa_url = (char*) malloc(url_len);
  sprintf(sa_url, "service:service-agent://%s", hostname);

  char* attrs = (char*) par("SAAttributes").stringValue();
  char* sa_attrs = expandVar(attrs, "$HOST", hostname);
  par("SAAttributes").setStringValue(sa_attrs);
  //free(attrs);

  agentService = new Service("service:service-agent", sa_url, 
			     sa_attrs, -1, simTime());

  serviceCache.setName("Service Cache");
  serviceCache.add(agentService);

  slpQueueLength.setName("SLPQueueLength");
}

void ServiceAgent::handleMessage ( cMessage * msg )
{

#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  printf("handleMessage('%s()')\n", msg->name()); 
#endif

  if (msg->isSelfMessage()) {
    
    if (strcmp(msg->name(), "ActiveDADiscovery") == 0) {
      handleActiveDADiscovery(check_and_cast<ServiceRequest*>(msg));
      return;
    }

    if (strcmp(msg->name(), "Timeout") == 0) {
      handleTimeout(msg);
      return;
    }

  } else if (msg->arrivedOn("fromApp")) {

    handleSLPReg(check_and_cast<SLPReg*>(msg));
    
  } else {
    
    Sockaddr_in* saddr = (Sockaddr_in*) msg->removeObject("recvFrom");

    SLPv2Message * slp = check_and_cast<SLPv2Message *>(msg);
    
    switch(slp->getFunctionId()) {
    case SRV_RQST:
      handleServiceRequest
	(check_and_cast<ServiceRequest*>(slp), saddr);
      break;
    case SRV_TYPE_RQST:
      handleServiceTypeRequest
	(check_and_cast<ServiceTypeRequest *>(slp), saddr);
      break;
    case ATTR_RQST:
      handleAttributeRequest
	(check_and_cast<AttributeRequest*>(slp), saddr);
      break;
    case SRV_ACK:
      handleServiceAcknowledgement
	(check_and_cast<ServiceAcknowledgement*>(slp), saddr);
      break;
    case DA_ADVERT:
      handleDAAdvertisement
	(check_and_cast<DAAdvertisement*>(slp), saddr);
      break;
    default:
      error("Unknown function id: '%s'", slp->getFunctionId());
    }
  }
}


void ServiceAgent::handleSLPReg ( SLPReg * slp_reg )
{
  slpHandle = slp_reg->getHSLP();
  const char* locale = slpHandle.pcLang;
  const char* srv_url = slp_reg->getPcSrvURL();
  int lifetime = slp_reg->getUsLifetime();
  const char* srv_type = slp_reg->getPcSrvType();
  const char* srv_attrs = slp_reg->getPcAttrs();

  Service * srv = (Service *) serviceCache[srv_url];
  if (srv == NULL) {
    // New service: add it to the cache
    srv = new Service(srv_type, srv_url, srv_attrs, lifetime, simTime());
    serviceCache.add(srv);
  } else {
    // Service already in cache: refresh it
    if (slp_reg->getFresh()) {
      srv->setAttributes(srv_attrs);
    }
    srv->setLifetime(lifetime);
    srv->refresh(simTime());
  }

  if (strlen(DAAddresses) == 0) {
    SLPRegReport * slp_reg_report = 
      new SLPRegReport("SLPRegReport", SRV_ACK);
    slp_reg_report->setTimestamp(slp_reg->timestamp());
    slp_reg_report->setContextPointer(slp_reg->contextPointer());
    slp_reg_report->setHSLP(slpHandle);
    slp_reg_report->setErrCode(SLP_OK);
    
    send(slp_reg_report, "toApp");

    delete slp_reg;
    return;
  }

  URLEntry url_entry;
  url_entry.lifetime = lifetime;
  url_entry.URL = srv_url;
  
  ServiceRegistration * srv_reg = 
    new ServiceRegistration("SrvReg", SRV_REG);
  srv_reg->setTimestamp(slp_reg->timestamp());
  srv_reg->setContextPointer(slp_reg->contextPointer());
  srv_reg->setVersion(2);
  srv_reg->setFunctionId(SRV_REG);
  srv_reg->setOverflow(false);
  srv_reg->setFresh(slp_reg->getFresh());
  srv_reg->setRequestMulticast(false);
  srv_reg->setXid(xid++);
  srv_reg->setLanguageTag(locale);
  srv_reg->setEntry(url_entry);
  srv_reg->setServiceType(srv_type);
  srv_reg->setScopeList(useScopes);
  srv_reg->setAttrList(srv_attrs);
  
  int bytes =
    16 +                      // SLPv2 header: "en" lang tag
     6 + strlen(srv_url) +     // URL Entry
     2 + strlen(srv_type) +    // Service Type
     2 + strlen(useScopes) +   // Scope list
     2 + strlen(srv_attrs) +   // Service Type
     2;                        // No authentication blocks
  srv_reg->setLength(bytes*8);
  
  Sockaddr_in* saddr = new Sockaddr_in("sendTo", "slp",DAAddresses);
  srv_reg->addObject(saddr);
  
#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>');
  debug(srv_reg, saddr);
#endif
  
  // Set retransmission timer
  scheduleTimer(srv_reg, datagramTimeout);
  
  send(srv_reg, "toNet");
  delete slp_reg;
}

void ServiceAgent::handleServiceAcknowledgement
( ServiceAcknowledgement* srv_ack, Sockaddr_in* saddr )
{
#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(srv_ack, saddr);
#endif  

  // Disable retranimission timer
  cancelTimer(srv_ack->getXid());
  
  SLPRegReport * slp_reg_report = new SLPRegReport("SLPRegReport", SRV_ACK);
  slp_reg_report->setTimestamp(srv_ack->timestamp());
  slp_reg_report->setContextPointer(srv_ack->contextPointer());
  slp_reg_report->setHSLP(slpHandle);
  slp_reg_report->setErrCode(SLP_OK);
  
  send(slp_reg_report, "toApp");
  
  delete srv_ack;
  delete saddr;
}


void ServiceAgent::handleServiceRequest
( ServiceRequest * srv_rqst, Sockaddr_in* saddr )
{
  const char* srv_rqst_scope = srv_rqst->getScopeList();
  const char* service_type = srv_rqst->getServiceType();
  const char* locale = srv_rqst->getLanguageTag();

  bool has_attr_list_ext = false;
  SLPv2Extension* ext = (SLPv2Extension*) srv_rqst->encapsulatedMsg();
  while (ext != NULL) {
    if (ext->getExtensionId() == ATTR_LIST_EXT) {
      has_attr_list_ext = true;
      break;
    }   
    SLPv2Extension * ext = (SLPv2Extension*) ext->encapsulatedMsg();
  }

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(srv_rqst, saddr);
#endif

  if (agentService->hasType(service_type)) {

    SLPv2Message* advert_rply = createAdvertisementReply(srv_rqst);

    saddr->setName("sendTo");
    advert_rply->addObject(saddr);
    
#ifdef SLPV2_DEBUG
  debug('>');
  debug(advert_rply, saddr);
#endif

    send(advert_rply, "toNet");

    delete srv_rqst;
    return;
  }

  ServiceReply * srv_rply = new ServiceReply("SrvRply", SRV_RPLY);
  srv_rply->setTimestamp(srv_rqst->timestamp());
  srv_rply->setContextPointer(srv_rqst->contextPointer());
  srv_rply->setVersion(2);
  srv_rply->setFunctionId(SRV_RPLY);
  srv_rply->setOverflow(false);
  srv_rply->setFresh(false);
  srv_rply->setRequestMulticast(false);
  srv_rply->setXid(srv_rqst->getXid());
  srv_rply->setLanguageTag(locale);
  srv_rply->setErrorCode(OK);
  srv_rply->setEntriesArraySize(serviceCache.items());

  int bytes = 
    14 + strlen(locale) + // SLPv2 header + language tag
     2 +                  // Error code
     2;                   // URL Entry count

  cMessage* last_extension = srv_rply;
  int exts_length = 0;

  // Look for active services in the cache
  int services_found = 0;
  for (int i=0; i<serviceCache.items(); i++) {
    Service * service = (Service*) serviceCache[i];
    if (service == NULL) {
      continue;
    }
    
    if (service->hasExpired(simTime())) {
      delete serviceCache.remove(i);
      continue;
    }

    if (service->hasType(service_type)) {
      simtime_t lifetime = service->getTTL(simTime());
      const char* url = service->getURL();
      const char* attrs = service->getAttributes();
      
      URLEntry url_entry;
      url_entry.lifetime = (int) lifetime;
      url_entry.URL = url;
      
      int srv_bytes = 6 + strlen(url);  // URL Entry
      
      // RFC3059: Attribute List Extension Support
      //
      AttributeListExtension* attr_list_ext = NULL;
      if (has_attr_list_ext) {
	attr_list_ext = 
	  new AttributeListExtension("AttrListExt", ATTR_LIST_EXT);
	attr_list_ext->setExtensionId(ATTR_LIST_EXT);
	attr_list_ext->setServiceURL(url);
	attr_list_ext->setAttributeList(attrs);

	int ext_length = 
	  5 +                  // Extension header
	  2 + strlen(url) +    // URL
	  2 + strlen(attrs);   // Attribute list
	attr_list_ext->setLength(ext_length);

	exts_length += ext_length;
      }
      
      if ((28 + bytes + srv_bytes + exts_length) > MTU) {
	srv_rply->setOverflow(true);
	if (attr_list_ext != NULL) {
	  delete attr_list_ext;
	}
	break;
      }

      if (attr_list_ext != NULL) {
	last_extension->encapsulate(attr_list_ext);
	last_extension = attr_list_ext;
	srv_bytes += attr_list_ext->length();
      }
      
      srv_rply->setEntries(services_found++, url_entry);
      bytes += srv_bytes;
    }
  }
  
  if ((services_found == 0) && (srv_rqst->getRequestMulticast())) {
    delete srv_rqst;
    delete srv_rply;
    delete saddr;
    return;
  }

  srv_rply->setEntriesArraySize(services_found);
  srv_rply->setLength(bytes*8);

  saddr->setName("sendTo");
  srv_rply->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug('>');
  debug(srv_rply, saddr);
#endif
  
  sendReply(srv_rply);
  
  delete srv_rqst;
}


SLPv2Message* ServiceAgent::createAdvertisementReply
( ServiceRequest* srv_rqst )
{
  const char* sa_url = agentService->getURL();
  const char* sa_attrs = agentService->getAttributes();
  const char* locale = srv_rqst->getLanguageTag();

  // Active SA Discovery
  //
  SAAdvertisement* sa_advert = 
    new SAAdvertisement("SAAdvert", SA_ADVERT);
  sa_advert->setTimestamp(srv_rqst->timestamp());
  sa_advert->setContextPointer(srv_rqst->contextPointer());
  sa_advert->setVersion(2);
  sa_advert->setFunctionId(SA_ADVERT);
  sa_advert->setOverflow(false);
  sa_advert->setFresh(false);
  sa_advert->setRequestMulticast(false);
  sa_advert->setXid(srv_rqst->getXid());
  sa_advert->setLanguageTag(locale);
  sa_advert->setURL(sa_url);
  sa_advert->setScopeList(useScopes);
  sa_advert->setAttrList(sa_attrs);

  int bytes = 
    14 + strlen(locale) +      // SLPv2 header: "en" language tag
     2 +                       // Error code
     4 +                       // DA Boot Timestamp
     2 + strlen(sa_url) +      // URL
     2 + strlen(useScopes) +   // Scope list
     2 + strlen(sa_attrs)+     // Attribute list
     3;                        // SPI Length + #Auth Blocks=0
  
  sa_advert->setLength(bytes*8);
    
  return sa_advert;
}


void ServiceAgent::handleServiceTypeRequest
( ServiceTypeRequest * srv_type_rqst, Sockaddr_in* saddr )
{
  const char* locale = srv_type_rqst->getLanguageTag();
  const char* scope_list = srv_type_rqst->getScopeList();

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(srv_type_rqst, saddr);
#endif

  int room_left = 255;
  char* service_types = (char*) malloc(room_left);
  service_types[0]='\0';
  room_left--;
  
  int types_found = 0;
  for (int i=0; i<serviceCache.items(); i++) {
    Service * service = (Service*) serviceCache[i];
    if (service == NULL) {
      continue;
    }

    if (service->hasExpired(simTime())) {
      delete serviceCache.remove(i);
      continue;
    }

    const char* type = service->getType();
    if (agentService->hasType(type) ||
	(strstr(service_types, type) != NULL)) {
      continue;
    }

    types_found++;
    int type_length = strlen(type);
    if (type_length >= room_left) {
      int add_len = type_length - room_left;
      service_types = (char*) 
	realloc(service_types, strlen(service_types) + add_len + 1);
    }

    if (types_found > 1) {
      strcat(service_types, ",");
      room_left--;
    }
    
    strcat(service_types, type);
    room_left -= type_length;
  }

  if ((types_found > 0) || (! srv_type_rqst->getRequestMulticast())) {

    ServiceTypeReply * srv_type_rply = 
      new ServiceTypeReply("SrvTypeRply", SRV_TYPE_RPLY);
    srv_type_rply->setTimestamp(srv_type_rqst->timestamp());
    srv_type_rply->setContextPointer(srv_type_rqst->contextPointer());
    srv_type_rply->setVersion(2);
    srv_type_rply->setFunctionId(SRV_TYPE_RPLY);
    srv_type_rply->setOverflow(false);
    srv_type_rply->setFresh(false);
    srv_type_rply->setRequestMulticast(false);
    srv_type_rply->setXid(srv_type_rqst->getXid());
    srv_type_rply->setLanguageTag(locale); 
    srv_type_rply->setErrorCode(OK);
    srv_type_rply->setSrvtypeList(service_types);

    int bytes = 
      14 + strlen(locale) +        // SLPv2 header + language tag
       2 +                         // Error code 
       2 + strlen(service_types);  // SrvtypesList
    srv_type_rply->setLength(bytes*8);
      
    saddr->setName("sendTo");
    srv_type_rply->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug('>');
  debug(srv_type_rply, saddr);
#endif

    sendReply(srv_type_rply);

  } else {

    delete saddr;

  }

  delete srv_type_rqst;
}

void ServiceAgent::handleAttributeRequest
( AttributeRequest * attr_rqst, Sockaddr_in* saddr )
{
  const char* url = attr_rqst->getURL();
  const char* scope_list = attr_rqst->getScopeList();

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(attr_rqst, saddr);
#endif

  AttributeReply * attr_rply = 
    new AttributeReply("AttrRply", ATTR_RPLY);
  attr_rply->setTimestamp(attr_rqst->timestamp());
  attr_rply->setContextPointer(attr_rqst->contextPointer());
  attr_rply->setVersion(2);
  attr_rply->setFunctionId(ATTR_RPLY);
  attr_rply->setOverflow(false);
  attr_rply->setFresh(false);
  attr_rply->setRequestMulticast(false);
  attr_rply->setXid(attr_rqst->getXid());
  attr_rply->setLanguageTag("en");

  int bytes = 
    16 +             // SLPv2 header: "en" lang tag
     2;              // Error code

  const char* attrs = NULL;
  Service * service = (Service *) serviceCache[url];
  if (service == NULL) {
    if (attr_rqst->getRequestMulticast()) {
      delete attr_rqst;
      delete attr_rply;
      delete saddr;
      return;
    }
    attr_rply->setErrorCode(INVALID_REGISTRATION);
    attr_rply->setAttrList(" ");
  } else {
    const char* attrs = service->getAttributes();
    attr_rply->setErrorCode(OK);
    attr_rply->setAttrList(attrs);

    bytes += 2 + strlen(attrs);  // Attr List
  }

  attr_rply->setLength(bytes*8);

  saddr->setName("sendTo");
  attr_rply->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug('>');
  debug(attr_rply, saddr);
#endif  

  sendReply(attr_rply);

  delete attr_rqst;
}


void ServiceAgent::sendReply ( SLPv2Message* msg )
{
  simtime_t processing_time = par("processingTime").doubleValue();
  if (occupiedUntil > simTime()) {
    simtime_t queue_time = occupiedUntil - simTime();
    slpQueueLength.record(queue_time);
    occupiedUntil += processing_time;
    processing_time += queue_time;
  } else {
    slpQueueLength.record(0.0);
    occupiedUntil = simTime() + processing_time;
  }

  sendDelayed(msg, processing_time, "toNet");
}


void ServiceAgent::handleFatalTimeout ( SLPv2Message * slpv2 )
{
  ServiceRequest* srv_rqst = dynamic_cast<ServiceRequest*>(slpv2);
  if ((srv_rqst != NULL) && 
      (strcmp(srv_rqst->getServiceType(), "directory-agent") == 0)) {
    
    if (strlen(useScopes) == 0) {
      useScopes = "DEFAULT";
      par("useScopes").setStringValue(useScopes);
    }
    
    // Active DA Discovery failed: re-schedule it
    //
    srv_rqst->setName("ActiveDADiscovery");
    simtime_t active_discovery_interval = 
      par("DAActiveDiscoveryInterval").doubleValue();
    scheduleAt(simTime() + active_discovery_interval, srv_rqst);
    return;
  }
  
  // Maximum retry time exceeded, sending SrvReg error to app
  //
  SLPRegReport * slp_reg_report = 
    new SLPRegReport("SLPRegReport", SRV_ACK);
  slp_reg_report->setTimestamp(slpv2->timestamp());
  slp_reg_report->setContextPointer(slpv2->contextPointer());
  slp_reg_report->setHSLP(slpHandle);
  slp_reg_report->setErrCode(SLP_NETWORK_TIMED_OUT);
  send(slp_reg_report, "toApp");
  
  delete slpv2;
}
