/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <omnetpp.h>
#include <string.h>

#include "UserAgent.h"
#include "Service.h"
#include "util.h"
#include "debug.h"

// simple UserAgent
//   parameters:
//     useScopes                 : string,
//     DAAddresses               : string,
//     DAActiveDiscoveryInterval : numeric,
//     multicastMaximuWait       : numeric,
//     multicastTimeout          : numeric,
//     DADiscoveryTimeout        : numeric,
//     datagramTimeout           : numeric,
//     ramdomWaitBound           : numeric,
//     MTU                       : numeric,
//     locale                    : string,
//     maxResults                : numeric;
// 
//   gates:
//      in: fromApp, fromNet;
//     out:   toApp,   toNet;
// endsimple

Define_Module ( UserAgent );

void UserAgent::initialize()
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("initialize()\n");
#endif

  useScopes = par("useScopes").stringValue();
  DAAddresses = par("DAAddresses").stringValue();
  locale = par("locale").stringValue();
  multicastMaximumWait = par("multicastMaximumWait").doubleValue();
  multicastTimeout = par("multicastTimeout").doubleValue();
  datagramTimeout = par("datagramTimeout").doubleValue();
  MTU = par("MTU");

  xid = 1;
  timers.setName("Timers");

  if (par("DAActiveDiscoveryInterval").doubleValue() > 0.0) {

    // Active DA Discovery enabled
    //
    const char* service_type = "service:directory-agent";
    ServiceRequest * srv_rqst = 
      new ServiceRequest("ActiveDADiscovery", SRV_RQST);
    srv_rqst->setVersion(2);
    srv_rqst->setFunctionId(SRV_RQST);
    srv_rqst->setOverflow(false);
    srv_rqst->setFresh(false);
    srv_rqst->setRequestMulticast(true);
    srv_rqst->setLanguageTag(locale);
    srv_rqst->setServiceType(service_type);

    int bytes = 
      14 + strlen(locale) +        // SLPv2 header + lang tag
      2 +                         // Empty PRList
      2 + strlen(service_type) +  // Service Type
      2 +                         // Emtpy search filter
      2;                          // No authentication blocks

    if (strlen(useScopes) == 0) {
      srv_rqst->setScopeList(" ");
      bytes += 2;
    } else {
      srv_rqst->setScopeList(useScopes);
      bytes += 2 + strlen(useScopes);
    }

    srv_rqst->setLength(bytes*8);
    
    simtime_t random_wait_bound = par("randomWaitBound").doubleValue();
    scheduleAt(simTime() + random_wait_bound, srv_rqst);

  } else if (strlen(useScopes) == 0) {

    useScopes = "DEFAULT";
    par("useScopes").setStringValue(useScopes);

  }
}

void UserAgent::handleMessage ( cMessage * msg )
{
#ifdef MOD_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), 'D');
  ev.printf("handleMessage('%s()')\n", msg->name());
#endif

  if (msg->isSelfMessage()) {
  
    if (strcmp(msg->name(), "ActiveDADiscovery") == 0) {
      handleActiveDADiscovery(check_and_cast<ServiceRequest*>(msg));
      return;
    }

    if (strcmp(msg->name(), "MulticastDiscoveryTimeout") == 0) {
      handleMulticastDiscoveryTimeout(msg);
      return;
    }

    if (strcmp(msg->name(), "Timeout") == 0) {
      handleTimeout(msg);
      return;
    }

  } else if (msg->arrivedOn("fromApp")) {

    SLPFindSrvs * slp_find_srvs = dynamic_cast<SLPFindSrvs*>(msg);
    if (slp_find_srvs != NULL) {
      handleSLPFindSrvs(slp_find_srvs);
      return;
    }

    SLPFindSrvsTypes * slp_find_srvs_types = 
      dynamic_cast<SLPFindSrvsTypes*>(msg);
    if (slp_find_srvs_types != NULL) {
      handleSLPFindSrvsTypes(slp_find_srvs_types);
      return;
    }

    SLPFindAttrs * slp_find_attrs = dynamic_cast<SLPFindAttrs*>(msg);
    if (slp_find_attrs != NULL) {
      handleSLPFindAttrs(slp_find_attrs);
      return;
    }

  } else {

    Sockaddr_in* saddr = (Sockaddr_in*) msg->removeObject("recvFrom");

    SLPv2Message * slp = check_and_cast<SLPv2Message*>(msg);
    switch(slp->getFunctionId()) {
    case SRV_RQST:
    case ATTR_RQST:
      // Broadcast requests from other UAs, drop them
      delete slp;
      break;
    case SRV_RPLY:
      handleServiceReply
	(check_and_cast<ServiceReply*>(slp), saddr);
      break;
    case SRV_TYPE_RPLY:
      handleServiceTypeReply
	(check_and_cast<ServiceTypeReply*>(slp), saddr);
      break;
    case ATTR_RPLY:
      handleAttributeReply
	(check_and_cast<AttributeReply*>(slp), saddr);
      break;
    case DA_ADVERT:
      handleDAAdvertisement
	(check_and_cast<DAAdvertisement*>(slp), saddr);
      break;
    case SA_ADVERT:
      handleSAAdvertisement
	(check_and_cast<SAAdvertisement*>(slp), saddr);
      break;
    default:
      error("Unknown SLPv2 function id: '%d'", slp->getFunctionId());
    }
  }
}

void UserAgent::handleSLPFindSrvs ( SLPFindSrvs * slp_find_srvs )
{
  slpHandle = slp_find_srvs->getHSLP();
  const char* locale = slpHandle.pcLang;
  const char* service_type = slp_find_srvs->getPcServiceType();
  const char* scope_list = slp_find_srvs->getPcScopeList();
  if (strcmp(scope_list, " ") == 0) {
    scope_list = useScopes;
  }

  ServiceRequest * srv_rqst = new ServiceRequest("SrvRqst", SRV_RQST);
  srv_rqst->setTimestamp(slp_find_srvs->timestamp());
  srv_rqst->setContextPointer(slp_find_srvs->contextPointer());
  srv_rqst->setVersion(2);
  srv_rqst->setFunctionId(SRV_RQST);
  srv_rqst->setOverflow(false);
  srv_rqst->setFresh(false);
  srv_rqst->setRequestMulticast(strlen(DAAddresses) == 0);
  srv_rqst->setXid(xid++);
  srv_rqst->setLanguageTag(locale);
  srv_rqst->setServiceType(service_type);
  srv_rqst->setScopeList(scope_list);

  int bytes = 
    14 + strlen(locale) +        // SLPv2 header + lang tag
     2 +                         // Empty PRList
     2 + strlen(service_type) +  // Service Type
     2 + strlen(scope_list) +    // Scope list
     2 +                         // Empty search filter
     2;                          // No authentication blocks
  srv_rqst->setLength(bytes*8);

  // RFC3059: Attribute List Extension Support
  //
  SLPFindSrvsAttrs * slp_find_srvs_attrs = 
    dynamic_cast<SLPFindSrvsAttrs*>(slp_find_srvs);

  if (slp_find_srvs_attrs != NULL) {
    AttributeListExtension* attr_list_ext = 
      new AttributeListExtension("AttrListExt", ATTR_LIST_EXT);
    attr_list_ext->setExtensionId(ATTR_LIST_EXT);
    attr_list_ext->setServiceURL(" ");
    attr_list_ext->setAttributeList(" ");

    int ext_length = 
      5 +  // Extension header
      2 +  // Empty URL
      2;   // Empty Attribute list
    attr_list_ext->setLength(ext_length);

    srv_rqst->encapsulate(attr_list_ext);
  }

  Sockaddr_in* saddr;
  if (srv_rqst->getRequestMulticast()) {

    saddr = new Sockaddr_in("sendTo", "slp", "*");

    // Set Multicast Discovery Timer
    SLPSrvURLCallback * srv_url_callback;
    if (slp_find_srvs_attrs == NULL) {
      srv_url_callback = new SLPSrvURLCallback("SLPSrvURLCallback", SRV_RPLY);
    } else {
      SLPSrvURLAttrsCallback * srv_url_attrs_callback = 
	new SLPSrvURLAttrsCallback("SLPSrvURLAttrsCallback", SRV_RPLY);
      srv_url_attrs_callback->setPcAttrList(" ");
      srv_url_callback = srv_url_attrs_callback;
    }
    srv_url_callback->setTimestamp(slp_find_srvs->timestamp());
    srv_url_callback->setContextPointer(slp_find_srvs->contextPointer());
    srv_url_callback->setHSLP(slpHandle);
    srv_url_callback->setPcSrvURL(" ");
    srv_url_callback->setSLifetime(0);
    srv_url_callback->setErrCode(SLP_LAST_CALL);

    cMessage* mcast_discovery_timer = new cMessage
      ("MulticastDiscoveryTimeout");
    
    mcast_discovery_timer->encapsulate(srv_url_callback);

    scheduleAt(simTime() + par("multicastTimeout").doubleValue(), 
	       mcast_discovery_timer);

  } else {
    saddr = new Sockaddr_in("sendTo", "slp", DAAddresses);
  }
  srv_rqst->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>'); 
  debug(srv_rqst, saddr);
#endif
    
  // Set retransmission timer
  scheduleTimer(srv_rqst, datagramTimeout);

  send(srv_rqst, "toNet");

  delete slp_find_srvs;
}


void UserAgent::handleServiceReply
( ServiceReply * srv_rply, Sockaddr_in* saddr )
{
  int url_entries = srv_rply->getEntriesArraySize();

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(srv_rply, saddr);
#endif

  // Disable retransmission timer
  cancelTimer(srv_rply->getXid());

  int i;
  for (i=0; i<url_entries; i++) {
    URLEntry entry = srv_rply->getEntries(i);
    const char* url = entry.URL;
    int lifetime = entry.lifetime;
    const char* attrs = NULL;
   
    SLPv2Extension * ext = 
      (SLPv2Extension*) srv_rply->encapsulatedMsg();
    while (ext != NULL) {
      if (ext->getExtensionId() == ATTR_LIST_EXT) {
	AttributeListExtension* attr_list_ext =
	  check_and_cast<AttributeListExtension*>(ext);
	const char* ext_url = attr_list_ext->getServiceURL();
	if (strcmp(url, ext_url) == 0) {
	  attrs = attr_list_ext->getAttributeList();
	  break;
	} else {
	  attrs = NULL;
	}
      }
      ext = (SLPv2Extension*) ext->encapsulatedMsg();
    }

    SLPSrvURLCallback * srv_url_callback;
    if (attrs == NULL) {
      srv_url_callback = new SLPSrvURLCallback("SLPSrvURLCallback", SRV_RPLY);
    } else {
      SLPSrvURLAttrsCallback * srv_url_attrs_callback = 
	new SLPSrvURLAttrsCallback("SLPSrvURLAttrsCallback", SRV_RPLY);
      srv_url_attrs_callback->setPcAttrList(attrs);
      srv_url_callback = srv_url_attrs_callback;
    }
    srv_url_callback->setTimestamp(srv_rply->timestamp());
    srv_url_callback->setContextPointer(srv_rply->contextPointer());
    srv_url_callback->setHSLP(slpHandle);
    srv_url_callback->setPcSrvURL(url);
    srv_url_callback->setSLifetime(lifetime);
    if ((url_entries == i+1) && (strlen(DAAddresses) != 0)) {
      srv_url_callback->setErrCode(SLP_LAST_CALL);
    } else {
      srv_url_callback->setErrCode(SLP_OK);
    }
    
    send(srv_url_callback, "toApp");
  }

  if (i == 0) {
    
    // Empty response
    //
    SLPSrvURLCallback * srv_url_callback = 
      new SLPSrvURLCallback("SLPSrvURLCallback", SRV_RPLY);
    srv_url_callback->setTimestamp(srv_rply->timestamp());
    srv_url_callback->setContextPointer(srv_rply->contextPointer());
    srv_url_callback->setHSLP(slpHandle);
    srv_url_callback->setPcSrvURL(" ");
    srv_url_callback->setSLifetime(0);
    srv_url_callback->setErrCode(SLP_LAST_CALL);

    send(srv_url_callback, "toApp");
  }

  delete srv_rply;
  delete saddr;
}

void UserAgent::handleSLPFindSrvsTypes 
( SLPFindSrvsTypes* slp_find_srvs_types )
{
  slpHandle = slp_find_srvs_types->getHSLP();
  const char* locale = slpHandle.pcLang;
  const char* scope_list = slp_find_srvs_types->getPcScopeList();
  if (strcmp(scope_list, " ") == 0) {
    scope_list = useScopes;
  }

  ServiceTypeRequest * srv_type_rqst = 
    new ServiceTypeRequest("SrvTypeRqst", SRV_TYPE_RQST);
  srv_type_rqst->setTimestamp(slp_find_srvs_types->timestamp());
  srv_type_rqst->setContextPointer
    (slp_find_srvs_types->contextPointer());
  srv_type_rqst->setVersion(2);
  srv_type_rqst->setFunctionId(SRV_TYPE_RQST);
  srv_type_rqst->setOverflow(false);
  srv_type_rqst->setFresh(false);
  srv_type_rqst->setRequestMulticast(strlen(DAAddresses) == 0);
  srv_type_rqst->setXid(xid++);
  srv_type_rqst->setLanguageTag(locale);
  srv_type_rqst->setScopeList(scope_list);

  int bytes = 
    14 + strlen(locale) +        // SLPv2 header + lang tag
     2 +                         // PRList: empty
     2 +                         // Naming Authority: empty 
     2 + strlen(scope_list);     // Scope list

  srv_type_rqst->setLength(bytes*8);

  Sockaddr_in* saddr;
  if (srv_type_rqst->getRequestMulticast()) {
    saddr = new Sockaddr_in("sendTo", "slp", "*");
  } else {
    saddr = new Sockaddr_in("sendTo", "slp", DAAddresses);
  }
  srv_type_rqst->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>');
  debug(srv_type_rqst, saddr);
#endif
    
  // Set retransmission timer
  scheduleTimer(srv_type_rqst, datagramTimeout);

  send(srv_type_rqst, "toNet");

  delete slp_find_srvs_types;
}


void UserAgent::handleServiceTypeReply
( ServiceTypeReply * srv_type_rply, Sockaddr_in* saddr )
{
  const char* srv_types = srv_type_rply->getSrvtypeList();

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(srv_type_rply, saddr);
#endif

  // Disable retransmission timer
  cancelTimer(srv_type_rply->getXid());

  SLPSrvTypeCallback * srv_type_callback =
    new SLPSrvTypeCallback("SLPSrvTypeCallback", SRV_TYPE_RPLY);
  srv_type_callback->setTimestamp(srv_type_rply->timestamp());
  srv_type_callback->setContextPointer
    (srv_type_rply->contextPointer());
  srv_type_callback->setHSLP(slpHandle);
  srv_type_callback->setPcSrvTypes(srv_types);
  srv_type_callback->setErrCode(SLP_LAST_CALL);
    
  send(srv_type_callback, "toApp");

  delete srv_type_rply;
  delete saddr;
}

void UserAgent::handleSLPFindAttrs ( SLPFindAttrs * slp_find_attrs )
{
  slpHandle = slp_find_attrs->getHSLP();
  const char* url = slp_find_attrs->getPcURL();
  const char* scope_list = slp_find_attrs->getPcScopeList();
  if (strcmp(scope_list, " ") == 0) {
    scope_list = useScopes;
  }
  const char* attr_ids = slp_find_attrs->getPcAttrIds();
  
  AttributeRequest * attr_rqst = 
    new AttributeRequest("AttrRqst", ATTR_RQST);
  attr_rqst->setTimestamp(slp_find_attrs->timestamp());
  attr_rqst->setContextPointer(slp_find_attrs->contextPointer());
  attr_rqst->setVersion(2);
  attr_rqst->setFunctionId(ATTR_RQST);
  attr_rqst->setOverflow(false);
  attr_rqst->setFresh(false);
  attr_rqst->setRequestMulticast(strlen(DAAddresses) == 0);
  attr_rqst->setXid(xid++);
  attr_rqst->setLanguageTag("en");
  attr_rqst->setURL(url);
  attr_rqst->setScopeList(scope_list);
  
  int bytes =
    16 +                      // SLPv2 header: "en" lang tag
     2 +                      // Empty PRList
     2 + strlen(url) +        // URL
     2 + strlen(scope_list) + // Scope list
     2 +                      // Tag list
     2;                       // No authentication blocks
  attr_rqst->setLength(bytes*8);

  Sockaddr_in* saddr;
  if (attr_rqst->getRequestMulticast()) {
    saddr = new Sockaddr_in("sendTo", "slp", "*");
  } else {
    saddr = new Sockaddr_in("sendTo", "slp", DAAddresses);
  }
  attr_rqst->addObject(saddr);
  
#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>');
  debug(attr_rqst, saddr);
#endif  
      
  // Set retransmission timer
  scheduleTimer(attr_rqst, datagramTimeout);

  send(attr_rqst, "toNet");

  delete slp_find_attrs;
}


void UserAgent::handleAttributeReply
( AttributeReply * attr_rply, Sockaddr_in* saddr )
{
#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(attr_rply, saddr);
#endif  

  // Disable retransmission timer
  cancelTimer(attr_rply->getXid());

  SLPAttrCallback * attr_callback = new SLPAttrCallback
    ("SLPAttrCallback", ATTR_RPLY);
  attr_callback->setTimestamp(attr_rply->timestamp());
  attr_callback->setContextPointer(attr_rply->contextPointer());
  attr_callback->setHSLP(slpHandle);
  attr_callback->setPcAttrList(attr_rply->getAttrList());
  attr_callback->setErrCode(SLP_LAST_CALL);
  
  send(attr_callback, "toApp");

  delete attr_rply;
  delete saddr;
}

void UserAgent::handleDAAdvertisement
( DAAdvertisement* da_advert, Sockaddr_in* saddr )
{
#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(da_advert, saddr);
#endif

  // Disable retransmission timer
  if (da_advert->getXid() != 0) {
    cancelTimer(da_advert->getXid());
  }

  char * da_addr = url2addr(da_advert->getURL());
  if (strcmp(DAAddresses, da_addr) != 0) {
    DAAddresses = da_addr;
    par("DAAddresses").setStringValue(DAAddresses);
  }

  if (strlen(useScopes) == 0) {
    useScopes = strdup(da_advert->getScopeList());
    par("useScopes").setStringValue(useScopes);
  }
  
  delete da_advert;
  delete saddr;
}

void UserAgent::handleSAAdvertisement
( SAAdvertisement* sa_advert, Sockaddr_in* saddr )
{
#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '<');
  debug(sa_advert, saddr);
#endif

  // Disable retransmission timer
  if (sa_advert->getXid() != 0) {
    cancelTimer(sa_advert->getXid());
  }

  if (strlen(useScopes) == 0) {
    useScopes = strdup(sa_advert->getScopeList());
    par("useScopes").setStringValue(useScopes);
  }
  
  delete sa_advert;
  delete saddr;
}

void UserAgent::handleActiveDADiscovery 
( ServiceRequest* active_discovery )
{
  if (strlen(DAAddresses) > 0) {
    delete active_discovery;
    return;
  }

  active_discovery->setXid(xid++);
  ServiceRequest* srv_rqst = 
    (ServiceRequest*) active_discovery->dup();
  srv_rqst->setName("SrvRqst");

  Sockaddr_in* saddr = new Sockaddr_in("sendTo", "slp", "*");
  srv_rqst->addObject(saddr);

#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '>');
  debug(srv_rqst, saddr);
#endif  

  // Set retransmission timer
  scheduleTimer(srv_rqst, datagramTimeout);

  send(srv_rqst, "toNet");
}

void UserAgent::scheduleTimer ( SLPv2Message* msg, simtime_t timeout )
{
  cMessage * timeout_msg = new cMessage("Timeout");
  timeout_msg->setKind(1);
  timeout_msg->setTimestamp(simTime());
  timeout_msg->encapsulate((SLPv2Message*) msg->dup());
  scheduleAt(simTime() + timeout, timeout_msg);
  timers.insert(timeout_msg);
}


void UserAgent::cancelTimer ( int xid )
{
  for (cQueue::Iterator iter(timers, true); !iter.end(); iter++) {
    cMessage * msg = (cMessage*) iter();
    SLPv2Message * slpv2 = (SLPv2Message*) msg->encapsulatedMsg();
    if (slpv2->getXid() == xid) {
      if (msg->isScheduled()) {
	cancelEvent(msg);
      }
      delete timers.remove(msg);
      break;
    }
  }
}

void UserAgent::handleMulticastDiscoveryTimeout ( cMessage* msg )
{
  cMessage* callback = msg->decapsulate();
  send(callback, "toApp");
}

void UserAgent::handleTimeout ( cMessage * timeout_msg )
{
  timers.remove(timeout_msg);

  SLPv2Message * slpv2 = (SLPv2Message*) timeout_msg->decapsulate();
  simtime_t wait = simTime() - timeout_msg->timestamp();
  if (wait > multicastMaximumWait) {
    handleFatalTimeout(slpv2);
    delete timeout_msg;
    return;
  }

  int retries = timeout_msg->kind() + 1;
  timeout_msg->setKind(retries);
  timeout_msg->encapsulate((SLPv2Message*) slpv2->dup());

  // Retry message
  simtime_t timeout = retries * datagramTimeout;
  scheduleAt(simTime() + timeout, timeout_msg);
  timers.insert(timeout_msg);
    
#ifdef SLPV2_DEBUG
  debug(simTime(), parentModule()->fullName(), fullName(), '!');
  debug(slpv2);
#endif
  
  send(slpv2, "toNet");
}

void UserAgent::handleFatalTimeout ( SLPv2Message* slpv2 )
{
  // Maximum retry time exceeded, sending error to app
  SLPSrvURLCallback * srv_url_callback;
  SLPAttrCallback * attr_callback;
  switch(slpv2->getFunctionId()) {
  case ATTR_RQST:
    attr_callback = new SLPAttrCallback("SLPAttrCallback", ATTR_RPLY);
    attr_callback->setTimestamp(slpv2->timestamp());
    attr_callback->setContextPointer(slpv2->contextPointer());
    attr_callback->setPcAttrList(" ");
    attr_callback->setErrCode(SLP_NETWORK_TIMED_OUT);
    send(attr_callback, "toApp");
  case SRV_RQST:
    ServiceRequest* srv_rqst = 
      check_and_cast<ServiceRequest*>(slpv2);
    
    if (strcmp(srv_rqst->getServiceType(), 
	       "service:directory-agent") == 0){
      
      // Active DA Discovery failed: re-schedule it
      //
      simtime_t active_discovery_interval = 
	par("DAActiveDiscoveryInterval").doubleValue();
      scheduleAt(simTime() + active_discovery_interval, srv_rqst);
      
      if (strlen(useScopes) == 0) {
	
	// Try Active SA discovery
	//
	const char* service_type = "service:service-agent";
	ServiceRequest * srv_rqst = 
	  new ServiceRequest("SrvRqst", SRV_RQST);
	srv_rqst->setVersion(2);
	srv_rqst->setFunctionId(SRV_RQST);
	srv_rqst->setOverflow(false);
	srv_rqst->setFresh(false);
	srv_rqst->setRequestMulticast(true);
	srv_rqst->setXid(xid++);
	srv_rqst->setLanguageTag(locale);
	srv_rqst->setServiceType(service_type);
	srv_rqst->setScopeList(" ");
	
	int bytes = 
	  14 + strlen(locale) +       // SLPv2 header + lang tag
	  2 +                         // Empty PRList
	  2 + strlen(service_type) +  // Service Type
	  2 +                         // Empty scope list
	  2 +                         // Emtpy search filter
	  2;                          // No authentication blocks
	srv_rqst->setLength(bytes*8);
	
	Sockaddr_in* saddr = new Sockaddr_in("sendTo", "slp", "*");
	srv_rqst->addObject(saddr);
	
#ifdef SLPV2_DEBUG
	debug(simTime(), parentModule()->fullName(), 
	      fullName(), '>');
	debug(srv_rqst, saddr);
#endif  
	
	// Set retransmission timer
	scheduleTimer(srv_rqst, datagramTimeout);
	
	send(srv_rqst, "toNet");
      }
      return;
      
    } else if (strcmp(srv_rqst->getServiceType(), 
		      "service:service-agent") == 0) {
      
      // Active SA Discovery also failed: use DEFAULT scope
      //
      useScopes = "DEFAULT";
      par("useScopes").setStringValue(useScopes);

    } else {
      srv_url_callback = 
	new SLPSrvURLCallback("SLPSrvURLCallback", SRV_RPLY);
      srv_url_callback->setTimestamp(slpv2->timestamp());
      srv_url_callback->setContextPointer(slpv2->contextPointer());
      srv_url_callback->setPcSrvURL(" ");
      srv_url_callback->setSLifetime(0);
      srv_url_callback->setErrCode(SLP_NETWORK_TIMED_OUT);
      send(srv_url_callback, "toApp");
      break;
    }
  }
  
  delete slpv2;
}
