/* SLPv2 model for Omnet++
 * Copyright (C) 2004  Manuel Urueña <muruenya@it.uc3m.es>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <omnetpp.h>

#include "debug.h"

void debugSLPHeader ( SLPv2Message* msg )
{
  ev.printf("%s(%2d", msg->name(), msg->getXid());

  if (msg->getOverflow() || msg->getFresh() 
      || msg->getRequestMulticast()) {
    ev.printf(" ");
  }

  if (msg->getOverflow()) {
    ev.printf("O");
  }
  if (msg->getFresh()) {
    ev.printf("F");
  }
  if (msg->getRequestMulticast()) {
    ev.printf("R");
  }

  const char* locale = msg->getLanguageTag();
  if (strcmp(locale, "en") != 0) {
    ev.printf(" '%s'", locale);
  }
}

void debugMsgTail ( cMessage* msg, Sockaddr_in* saddr )
{
  if (saddr == NULL) {
    saddr = (Sockaddr_in*) msg->getObject("sendTo");
  }

  if (saddr == NULL) {
    saddr = (Sockaddr_in*) msg->getObject("recvFrom");
  }

  
  if (strcmp(saddr->name(), "sendTo") == 0) {
    ev.printf(") -> ");
  } else if (strcmp(saddr->name(), "recvFrom") == 0) {
    ev.printf(") <- ");
  } else {
    ev.printf(")");
  }
  
  if (saddr != NULL) {
    ev.printf("%s:%s", saddr->getAddress(), saddr->getPort());
  }

  ev.printf("    (%d bytes)\n", msg->length()/8);
}

const char* error(int err_code)
{
  char* err_msg;

  switch(err_code) {
  case OK: 
    err_msg = "OK"; 
    break;
  case LANGUAGE_NOT_SUPPORTED: 
    err_msg = "LANGUAGE_NOT_SUPPORTED"; 
    break;
  case PARSE_ERROR: 
    err_msg = "PARSE_ERROR"; 
    break;
  case INVALID_REGISTRATION:
    err_msg = "INVALID_REGISTRATION"; 
    break;
  case SCOPE_NOT_SUPPORTED:
    err_msg = "SCOPE_NOT_SUPPORTED"; 
    break;
  case VER_NOT_SUPPORTED:
    err_msg = "VER_NOT_SUPPORTED"; 
    break;
  case INTERNAL_ERROR:
    err_msg = "INTERNAL_ERROR"; 
    break;
  case DA_BUSY_NOW:
    err_msg = "DA_BUSY_NOW"; 
    break;
  case OPTION_NOT_UNDERSTOOD:
    err_msg = "OPTION_NOT_UNDERSTOOD"; 
    break;
  case INVALID_UPDATE:
    err_msg = "INVALID_UPDATE"; 
    break;
  case MSG_NOT_SUPPORTED:
    err_msg = "MSG_NOT_SUPPORTED"; 
    break;
  case REFRESH_REJECTED:
    err_msg = "REFRESH_REJECTED"; 
    break;
  default:
    err_msg = (char*) malloc(255);
    sprintf(err_msg, "%d", err_code);
  }

  return err_msg;
}


void debug 
( simtime_t time, const char* host, const char* app, char type )
{
  if (app != NULL) {
    ev.printf("%.3f %5s.%-6s \t%c ", time, host, app, type);
  } else {
    ev.printf("%.3f %5s        \t%c ", time, host, type);
  }
}


void debug ( char type )
{
  ev.printf("\t\t\t%c ", type);
}

void debug ( SLPFindSrvs* slp_find_srvs )
{
  ev << slp_find_srvs->name() << "('"
     << slp_find_srvs->getPcServiceType() << "'";
  if (strcmp(slp_find_srvs->getPcScopeList(), " ") != 0) {
    ev << ", '" << slp_find_srvs->getPcScopeList() << "'";
  }
  ev << ")" << endl;
}

void debug ( SLPSrvURLAttrsCallback* callback )
{
  ev.printf("%s('%.25s' (", callback->name(), callback->getPcSrvURL());
  ev << callback->getSLifetime() << "), '";
  ev.printf("%.30s', ", callback->getPcAttrList());
  if (callback->getErrCode() == SLP_LAST_CALL) {
    ev << "SLP_LAST_CALL";
  } else if (callback->getErrCode() == SLP_OK) {
    ev << "SLP_OK";
  } else {
    ev << callback->getErrCode();
  }
  ev << ")" << endl;
}

void debug ( SLPSrvURLCallback* callback )
{
  ev.printf("%s(", callback->name());
  if (strcmp(callback->getPcSrvURL(), " ") != 0) {
    ev.printf("'%.25s' (%d), ", callback->getPcSrvURL(), 
	      callback->getSLifetime());
  }
  if (callback->getErrCode() == SLP_NETWORK_TIMED_OUT) {
    ev << "SLP_NETWORK_TIMED_OUT)";
  } else if (callback->getErrCode() == SLP_LAST_CALL) {
    ev << "SLP_LAST_CALL";
  } else if (callback->getErrCode() == SLP_OK) {
    ev << "SLP_OK";
  } else {
    ev << callback->getErrCode();
  }
  ev << ")" << endl;
}

void debug ( SLPFindAttrs* slp_find_attrs )
{
  ev.printf("%s('%.30s'", slp_find_attrs->name(), 
	    slp_find_attrs->getPcURL());
  if (strcmp(slp_find_attrs->getPcScopeList(), " ") != 0) {
    ev << ", '" << slp_find_attrs->getPcScopeList() << "'";
  }
  ev << ")" << endl;
}

void debug ( SLPAttrCallback* callback )
{
  ev.printf("%s(", callback->name());
  if (strcmp(callback->getPcAttrList(), " ") != 0) {
    ev.printf("'%.30s', ", callback->getPcAttrList());
  }
  if (callback->getErrCode() == SLP_NETWORK_TIMED_OUT) {
    ev << "SLP_NETWORK_TIMED_OUT)";
  } else if (callback->getErrCode() == SLP_LAST_CALL) {
    ev << "SLP_LAST_CALL";
  } else if (callback->getErrCode() == SLP_OK) {
    ev << "SLP_OK";
  } else {
    ev << callback->getErrCode();
  }
  SLPSrvURLCallback * srv_url_callback = 
    (SLPSrvURLCallback *) callback->contextPointer();
  ev.printf(" ['%.25s']\n", srv_url_callback->getPcSrvURL());
}

void debug ( SLPFindSrvsTypes* slp_find_srvs_types )
{
  ev << slp_find_srvs_types->name() << "(";
  if (strcmp(slp_find_srvs_types->getPcScopeList(), " ") != 0) {
    ev << "'" << slp_find_srvs_types->getPcScopeList() << "'";
  }
  ev << ")" << endl;
}

void debug ( SLPSrvTypeCallback * callback )
{
  ev.printf("%s(", callback->name());
  if (strcmp(callback->getPcSrvTypes(), " ") != 0) {
    ev.printf("'%s', ", callback->getPcSrvTypes());
  }
  if (callback->getErrCode() == SLP_NETWORK_TIMED_OUT) {
    ev << "SLP_NETWORK_TIMED_OUT)";
  } else if (callback->getErrCode() == SLP_LAST_CALL) {
    ev << "SLP_LAST_CALL";
  } else if (callback->getErrCode() == SLP_OK) {
    ev << "SLP_OK";
  } else {
    ev << callback->getErrCode();
  }
  ev << ")" << endl;
}

void debug 
( SLPv2Message* slp, Sockaddr_in* saddr )
{
  switch(slp->getFunctionId()) {
  case SRV_RQST:
    debug(check_and_cast<ServiceRequest*>(slp), saddr);
    break;
  case SRV_RPLY:
    debug(check_and_cast<ServiceReply*>(slp), saddr);
    break;
  case SRV_REG:
    debug(check_and_cast<ServiceRegistration*>(slp), saddr);
    break;
  case SRV_ACK:
    debug(check_and_cast<ServiceAcknowledgement*>(slp), saddr);
    break;
  case DA_ADVERT:
    debug(check_and_cast<DAAdvertisement*>(slp), saddr);
    break;
  case SA_ADVERT:
    debug(check_and_cast<SAAdvertisement*>(slp), saddr);
    break;
  case SRV_TYPE_RQST:
    debug(check_and_cast<ServiceTypeRequest*>(slp), saddr);
    break;
  case SRV_TYPE_RPLY:
    debug(check_and_cast<ServiceTypeReply*>(slp), saddr);
    break;
  case ATTR_RQST:
    debug(check_and_cast<AttributeRequest*>(slp), saddr);
    break;
  case ATTR_RPLY:
    debug(check_and_cast<AttributeReply*>(slp), saddr);
    break;
  case SRV_DEREG:
    debug(check_and_cast<ServiceDeregistration*>(slp), saddr);
    break;
  }
}

void debug 
( ServiceRequest* srv_rqst, Sockaddr_in* saddr )
{
  debugSLPHeader(srv_rqst);
  ev.printf(", '%s'", srv_rqst->getServiceType());

  const char* scope_list = srv_rqst->getScopeList();
  if (strcmp(scope_list, " ") != 0) {
    ev.printf(", '%s'", scope_list);
  }

  SLPv2Extension * ext = 
    (SLPv2Extension*) srv_rqst->encapsulatedMsg();
  while (ext != NULL) {
    if (ext->getExtensionId() == ATTR_LIST_EXT) {
      ev.printf(", [AttrListExt]");
      break;
    }
    ext = (SLPv2Extension*) ext->encapsulatedMsg();
  }

  debugMsgTail(srv_rqst, saddr);
}


void debug 
( ServiceReply* srv_rply, Sockaddr_in* saddr )
{
  debugSLPHeader(srv_rply);

  int err_code = srv_rply->getErrorCode();
  if (err_code != OK) {
    ev.printf(", %s", error(err_code));
  }

  int url_entries = srv_rply->getEntriesArraySize();
  for (int i=0; i<url_entries; i++) {

    URLEntry entry = srv_rply->getEntries(i);
    const char* url = entry.URL;
    int lifetime = entry.lifetime;
    const char* attrs = NULL;

    SLPv2Extension * ext = 
      (SLPv2Extension*) srv_rply->encapsulatedMsg();
    while (ext != NULL) {
      if (ext->getExtensionId() == ATTR_LIST_EXT) {
	AttributeListExtension* attr_list_ext =
	  check_and_cast<AttributeListExtension*>(ext);
	const char* ext_url = attr_list_ext->getServiceURL();
	if (strcmp(url, ext_url) == 0) {
	  attrs = attr_list_ext->getAttributeList();
	  break;
	} else {
	  attrs = NULL;
	}
      }
      ext = (SLPv2Extension*) ext->encapsulatedMsg();
    }

#ifdef SLPV2_DEBUG
    ev.printf(", ");
    if (i > 0) {
      ev.printf("\n\t\t\t| \t      ");
    }
    ev.printf("'%.25s' (%d)", url, lifetime);
    if (attrs != NULL) {
      ev.printf("+[AttrListExt:'%.30s']", attrs);
    }
#endif
  }
    
  debugMsgTail(srv_rply, saddr);
}


void debug 
( ServiceRegistration* srv_reg, Sockaddr_in* saddr )
{
  debugSLPHeader(srv_reg);

  URLEntry entry = srv_reg->getEntry();
  int lifetime = entry.lifetime;
  const char* url = entry.URL;

  ev.printf(", '%.25s' (%d), '%s', '%s'", url, lifetime, 
	    srv_reg->getServiceType(), srv_reg->getScopeList());
  
  const char* attr_list = srv_reg->getAttrList();
  if (strcmp(attr_list, " ") != 0) {
    ev.printf(", '%.30s'", attr_list);
  }
    
  debugMsgTail(srv_reg, saddr);
}


void debug 
( ServiceAcknowledgement* srv_ack, Sockaddr_in* saddr )
{
  debugSLPHeader(srv_ack);

  int err_code = srv_ack->getErrorCode();
  ev.printf(", %s", error(err_code));

  debugMsgTail(srv_ack, saddr);
}


void debug 
( DAAdvertisement* da_advert, Sockaddr_in* saddr )
{
  debugSLPHeader(da_advert);

  int err_code = da_advert->getErrorCode();
  if (err_code != OK) {
    ev.printf(", %s", error(err_code));
  }

  ev.printf(", %ld, '%.25s', '%s'", da_advert->getBootTimestamp(),
	    da_advert->getURL(), da_advert->getScopeList());

  const char* attr_list = da_advert->getAttrList();
  if (strcmp(attr_list, " ") != 0) {
    ev.printf(", '%.30s'", attr_list);
  }

  debugMsgTail(da_advert, saddr);
}


void debug 
( SAAdvertisement* sa_advert, Sockaddr_in* saddr )
{
  debugSLPHeader(sa_advert);

  ev.printf(", '%.25s', '%s'", sa_advert->getURL(), 
	    sa_advert->getScopeList());

  const char* attr_list = sa_advert->getAttrList();
  if (strcmp(attr_list, " ") != 0) {
    ev.printf(", '%.30s'", attr_list);
  }

  debugMsgTail(sa_advert, saddr);
}


void debug 
( ServiceTypeRequest* srv_type_rqst, Sockaddr_in* saddr )
{
  debugSLPHeader(srv_type_rqst);

  ev.printf(", '%s'", srv_type_rqst->getScopeList());

  debugMsgTail(srv_type_rqst, saddr);
}


void debug 
( ServiceTypeReply* srv_type_rply, Sockaddr_in* saddr )
{
  debugSLPHeader(srv_type_rply);

  int err_code = srv_type_rply->getErrorCode();
  if (err_code != OK) {
    ev.printf(", %s", error(err_code));
  }
  
  const char* types = srv_type_rply->getSrvtypeList();
  if (strcmp(types, " ") == 0) {
    ev.printf(", ''");
  } else {
    ev.printf(", '%s'", types);
  }

  debugMsgTail(srv_type_rply, saddr);
}


void debug 
( AttributeRequest* attr_rqst, Sockaddr_in* saddr )
{
  debugSLPHeader(attr_rqst);

  ev.printf(", '%.25s', '%.30s'", attr_rqst->getURL(), 
	    attr_rqst->getScopeList());

  debugMsgTail(attr_rqst, saddr);
}


void debug 
( AttributeReply* attr_rply, Sockaddr_in* saddr )
{
  debugSLPHeader(attr_rply);

  int err_code = attr_rply->getErrorCode();
  if (err_code != OK) {
    ev.printf(", %s", error(err_code));
  }
  
  const char* attr_list = attr_rply->getAttrList();
  if (strcmp(attr_list, " ") == 0) {
    ev.printf(", ''");
  } else {
    ev.printf(", '%.30s'", attr_list);
  }

  debugMsgTail(attr_rply, saddr);
}

void debug 
( ServiceDeregistration* srv_dereg, Sockaddr_in* saddr )
{
  debugSLPHeader(srv_dereg);

  ev.printf(", '%s", srv_dereg->getScopeList());

  URLEntry entry = srv_dereg->getEntry();
  int lifetime = entry.lifetime;
  const char* url = entry.URL;

  ev.printf("'%.25s' (%d)", url, lifetime);
  
  debugMsgTail(srv_dereg, saddr);
}

void debug 
( ClientRequest* clnt_rqst, Sockaddr_in* saddr )
{
  ev.printf("%s(%2d", clnt_rqst->name(), clnt_rqst->getId());

  debugMsgTail(clnt_rqst, saddr);
}

void debug 
( ServerResponse* srv_rsp, Sockaddr_in* saddr )
{
  ev.printf("%s(%2d", srv_rsp->name(), srv_rsp->getId());

  debugMsgTail(srv_rsp, saddr);
}
